unit LowLevelGPIO;

{$mode objfpc}{$H+}

interface

uses Classes, SysUtils;

const
  REG_GPIO = $3F000;
  PAGE_SIZE = 4096;
  BLOCK_SIZE = 4096;
  CLOCK_BASE = (REG_GPIO + $101);
  GPIO_BASE = (REG_GPIO + $200);
  GPIO_PWM = (REG_GPIO + $20C);

  INPUT = 0;
  OUTPUT = 1;
  PWM_OUTPUT = 2;
  LOW = False;
  HIGH = True;
  PUD_OFF = 0;
  PUD_DOWN = 1;
  PUD_UP = 2;

  PWM_CONTROL = 0;
  PWM_STATUS = 4;
  PWM0_RANGE = 16;
  PWM0_DATA = 20;
  PWM1_RANGE = 32;
  PWM1_DATA = 36;

  PWMCLK_CNTL = 160;
  PWMCLK_DIV = 164;

  PWM1_MS_MODE = $8000;
  PWM1_USEFIFO = $2000;
  PWM1_REVPOLAR = $1000;
  PWM1_OFFSTATE = $0800;
  PWM1_REPEATFF = $0400;
  PWM1_SERIAL = $0200;
  PWM1_ENABLE = $0100;

  PWM0_MS_MODE = $0080;
  PWM0_USEFIFO = $0020;
  PWM0_REVPOLAR = $0010;
  PWM0_OFFSTATE = $0008;
  PWM0_REPEATFF = $0004;
  PWM0_SERIAL = $0002;
  PWM0_ENABLE = $0001;

type
  TIoPort = class
  private
  public
    FGpio: ^longword;
    FClk: ^longword;
    FPwm: ^longword;
    procedure SetPinMode(gpin, mode: byte);
    function GetBit(gpin: byte): boolean;
    procedure ClearBit(gpin: byte);
    procedure SetBit(gpin: byte);
    procedure SetPullMode(gpin, mode: byte);
    procedure PwmWrite(gpin: byte; Value: longword);
  end;

  TIoDriver = class
  private
  public
    destructor Destroy; override;
    function MapIo: boolean;
    procedure UnmapIoRegisrty(FMap: TIoPort);
    function CreatePort(PortGpio, PortClk, PortPwm: longword): TIoPort;
  end;

var fd: integer;

procedure DelayNanoseconds(howLong: longword);

implementation

uses baseUnix, Unix;

procedure DelayNanoseconds(howLong: longword);
var sleeper,dummy: timespec;
begin
  sleeper.tv_sec := 0;
  sleeper.tv_nsec := howLong;
  fpnanosleep(@sleeper,@dummy);
end;

destructor TIoDriver.Destroy;
begin
  inherited Destroy;
end;

function TIoDriver.MapIo: boolean;
begin
  result := true;
  fd := fpopen('/dev/mem', O_RdWr or O_Sync);
  if fd < 0 then result := false;
end;

procedure TIoDriver.UnmapIoRegisrty(FMap: TIoPort);
begin
  if FMap.FGpio <> nil then
    begin
      fpMUnmap(FMap.FGpio, PAGE_SIZE);
      FMap.FGpio := nil;
    end;
  if FMap.FClk <> nil then
    begin
      fpMUnmap(FMap.FClk, PAGE_SIZE);
      FMap.FClk := nil;
    end;
  if FMap.FPwm <> nil then
    begin
      fpMUnmap(FMap.FPwm, PAGE_SIZE);
      FMap.FPwm := nil;
    end;
end;

function TIoDriver.CreatePort(PortGpio,PortClk,PortPwm: longword): TIoPort;
begin
  Result := TIoPort.Create;
  Result.FGpio := FpMmap(nil,PAGE_SIZE,PROT_READ or PROT_WRITE,MAP_SHARED,fd,PortGpio);
  Result.FClk := FpMmap(nil,PAGE_SIZE,PROT_READ or PROT_WRITE,MAP_SHARED,fd,PortClk);
  Result.FPwm := FpMmap(nil,PAGE_SIZE,PROT_READ or PROT_WRITE,MAP_SHARED,fd,PortPwm);
end;

procedure TIoPort.SetPinMode(gpin, mode: byte);
var fSel,shift,alt: byte;
    gpiof,clkf,pwmf: ^longword;
begin
  fSel := (gpin div 10) * 4;
  shift := (gpin mod 10) * 3;
  gpiof := pointer(longword(Self.FGpio) + fSel);
  if (mode = INPUT) then
    gpiof^ := gpiof^ and ($FFFFFFFF - (7 shl shift))
  else if (mode = OUTPUT) then
    gpiof^ := gpiof^ and ($FFFFFFFF - (7 shl shift)) or (1 shl shift)
  else if (mode = PWM_OUTPUT) then
    begin
      case gpin of
        12, 13, 40, 41, 45: alt := 4;
        18, 19: alt := 2;
        else alt := 0;
      end;
    if alt > 0 then
      begin
        gpiof^ := gpiof^ and ($FFFFFFFF - (7 shl shift)) or (alt shl shift);
        clkf := pointer(longword(Self.FClk) + PWMCLK_CNTL);
        clkf^ := $5A000011 or (1 shl 5);
        DelayNanoseconds(200);
        clkf := pointer(longword(Self.FClk) + PWMCLK_DIV);
        clkf^ := $5A000000 or (32 shl 12);
        clkf := pointer(longword(Self.FClk) + PWMCLK_CNTL);
        clkf^ := $5A000011;
        Self.ClearBit(gpin);
        pwmf := pointer(longword(Self.FPwm) + PWM_CONTROL);
        pwmf^ := 0;
        DelayNanoseconds(200);
        pwmf := pointer(longword(Self.FPwm) + PWM0_RANGE);
        pwmf^ := $400;
        DelayNanoseconds(200);
        pwmf := pointer(longword(Self.FPwm) + PWM1_RANGE);
        pwmf^ := $400;
        DelayNanoseconds(200);
        pwmf := pointer(longword(Self.FPwm) + PWM0_DATA);
        pwmf^ := 0;
        pwmf := pointer(longword(Self.FPwm) + PWM1_DATA);
        pwmf^ := 0;
        pwmf := pointer(longword(Self.FPwm) + PWM_CONTROL);
        pwmf^ := PWM0_ENABLE or PWM1_ENABLE;
      end;
    end;
end;

procedure TIoPort.SetBit(gpin: byte);
var gpiof: ^longword;
begin
  gpiof := pointer(longword(Self.FGpio) + 28 + (gpin shr 5) shl 2);
  gpiof^ := 1 shl gpin;
end;

procedure TIoPort.ClearBit(gpin: byte);
var gpiof: ^longword;
begin
  gpiof := pointer(longword(Self.FGpio) + 40 + (gpin shr 5) shl 2);
  gpiof^ := 1 shl gpin;
end;

function TIoPort.GetBit(gpin: byte): boolean;
var gpiof: ^longword;
begin
  gpiof := pointer(longword(Self.FGpio) + 52 + (gpin shr 5) shl 2);
  if (gpiof^ and (1 shl gpin)) = 0 then
    result := false
  else result := true;
end;

procedure TIoPort.SetPullMode(gpin, mode: byte);
var pudf, pudclkf: ^longword;
begin
  pudf := pointer(longword(Self.FGpio) + 148);
  pudf^ := mode;
  DelayNanoseconds(200);
  pudclkf := pointer(longword(Self.FGpio) + 152 + (gpin shr 5) shl 2);
  pudclkf^ := 1 shl gpin;
  DelayNanoseconds(200);
  pudf^ := 0;
  pudclkf^ := 0;
end;

procedure TIoPort.PwmWrite(gpin: byte; Value: longword);
var pwmf: ^longword;
    port: byte;
begin
  case gpin of
    12, 18, 40: port := PWM0_DATA;
    13, 19, 41, 45: port := PWM1_DATA;
    else exit;
  end;
  pwmf := pointer(longword(self.FPwm) + port);
  pwmf^ := Value and $FFFFFBFF;
end;


end.
